#include <stdint.h>
#include <stdarg.h>
#include <stdlib.h>

#include "debug.h"
#include "proto.h"

#define MARGIN_X 40
#define MARGIN_Y 40

#define DDR_SIZE (16*1024*1024)

#define L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_MODE_SET    0x100
#define L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_SYNC        0x101
#define L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_FLIP        0x102

#define L1GPU_DISPLAY_SYNC_HSYNC                    1
#define L1GPU_DISPLAY_SYNC_VSYNC                    2


static uint64_t memory_handle, context_handle;
static uint64_t ddr_lpar;

static int posx, posy;
static unsigned int fb_fg = 0x00ffffff;
static unsigned int fb_bg = 0;

void *fb_addr = 0;


void draw_glyph_8x8(int px, int py, const unsigned char *g,
		    unsigned int fg, unsigned int bg)
{
  uint32_t *fb = (uint32_t *)(fb_addr + px*4 + py * av_pitch);
  int i, j;
  for(i=0; i<8; i++) {
    uint8_t m = *g++;
    for(j=0; j<8; j++) {
      fb[j] = ((m&0x80)? fg : bg);
      m <<= 1;
    }
    fb += av_pitch/sizeof(uint32_t);
  }
}

void fb_gotoxy(int x, int y)
{
  posx=x;
  posy=y;
}

void fb_setcolor(unsigned int fg, unsigned int bg)
{
  fb_fg=fg;
  fb_bg=bg;
}

void fb_putc(char c)
{
  switch(c) {
  case '\n':
    ++posy;
  case '\r':
    posx=0;
    break;
  case '\f':
    posx=posy=0;
    uint64_t *p = fb_addr, v = (((uint64_t)fb_bg)<<32)|fb_bg;
    int i;
    for(i=0; i<(av_pitch*av_height)/sizeof(uint64_t); i++)
      *p++ = v;
    break;
  default:
    draw_glyph_8x8(posx*8+MARGIN_X, posy*8+MARGIN_Y, font8x8+c*8,
		   fb_fg, fb_bg);
    if(++posx >= (av_width-2*MARGIN_X)/8) {
      posx=0;
      ++posy;
    }
  }
}

void fb_puts(const char *s)
{
  while(*s)
    fb_putc(*s++);
}

void fb_vprintf(const char *fmt, va_list ap)
{
  char *s = 0;
  int i, l = vasprintf(&s, fmt, ap);
  fb_puts(s);
  free(s);
}

void fb_printf(const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  fb_vprintf(fmt, ap);
  va_end(ap);
}

void fb_clear(unsigned int colour)
{
  uint64_t *p = fb_addr, v = (((uint64_t)colour)<<32)|colour;
  int i;
  for(i=0; i<DDR_SIZE/sizeof(uint64_t); i++)
    *p++ = v;
}

void fb_flip(unsigned long offset)
{
  lv1_gpu_context_attribute(context_handle,
			    L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_FLIP,
			    0, offset, 0, 0);
  lv1_gpu_context_attribute(context_handle,
			    L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_FLIP,
			    1, offset, 0, 0);
}

void fb_init()
{
  long status;
  uint64_t lpar_dma_control, lpar_driver_info, lpar_reports, lpar_reports_size;

  DBG("lv1_gpu_context_attribute(0x0, L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_MODE_SET, 0, 0, 0, 0)");
  status = lv1_gpu_context_attribute(0x0, 
				     L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_MODE_SET,
				     0, 0, 0, 0);	// head a
  DBG("=> %ld\n", status);

  DBG("lv1_gpu_context_attribute(0x0, L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_MODE_SET, 0, 0, 1, 0)");
  status = lv1_gpu_context_attribute(0x0, 
				     L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_MODE_SET,
				     0, 0, 1, 0);	// head b
  DBG("=> %ld\n", status);
  
  DBG("lv1_gpu_context_attribute(0x0, L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_SYNC, 0, L1GPU_DISPLAY_SYNC_VSYNC, 0, 0)");
  status = lv1_gpu_context_attribute(0x0, 
				     L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_SYNC,
				     0, L1GPU_DISPLAY_SYNC_VSYNC, 0, 0);
  DBG("=> %ld\n", status);

  DBG("lv1_gpu_context_attribute(0x0, L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_SYNC, 0, L1GPU_DISPLAY_SYNC_VSYNC, 0, 0)");
  status = lv1_gpu_context_attribute(0x0, 
				     L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_SYNC,
				     1, L1GPU_DISPLAY_SYNC_VSYNC, 0, 0);
  DBG("=> %ld\n", status);

  DBG("lv1_gpu_memory_allocate(%ld, 0, 0, 0, 0, %p, %p)",
      DDR_SIZE, &memory_handle, &ddr_lpar);
  status = lv1_gpu_memory_allocate(DDR_SIZE,
				   0, 0, 0, 0,
				   &memory_handle, &ddr_lpar);
  DBG("=> %ld\n", status);
  DBG("ddr_lpar = %p\n", ddr_lpar);

  fb_addr = ioremap(ddr_lpar, DDR_SIZE);
  fb_clear(0);
  DBG("fb_addr = %p\n", fb_addr);

  DBG("lv1_gpu_context_allocate(%p, 0, %p, %p, %p, %p, %p)",
      memory_handle, 0, &context_handle, &lpar_dma_control, &lpar_driver_info, 
      &lpar_reports, &lpar_reports_size);
  status = lv1_gpu_context_allocate(memory_handle, 0,
				    &context_handle, 
				    &lpar_dma_control, &lpar_driver_info, 
				    &lpar_reports, &lpar_reports_size);
  DBG("=> %ld\n", status);

  fb_flip(0);
}

void fb_term()
{
  fb_addr = 0;
  lv1_gpu_context_free(context_handle);
  lv1_gpu_memory_free(memory_handle);
}
