#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include <malloc.h>

#include "debug.h"
#include "proto.h"

#include "zlib.h"

#define SCRATCHPAD_ADDR (123*1024*1024+0xfe000)
#define DEVTREE_ADDR    (123*1024*1024+0xff000)

static void *kernel = NULL, *initrd = NULL;
static unsigned long kernel_size, initrd_size, kernel_memsize;

typedef uint16_t Elf64_Half;
typedef uint32_t Elf64_Word;
typedef uint64_t Elf64_Xword;
typedef uint64_t Elf64_Addr;
typedef uint64_t Elf64_Off;

#define EI_NIDENT (16)
typedef struct
{
  unsigned char	e_ident[EI_NIDENT];	/* Magic number and other info */
  Elf64_Half	e_type;			/* Object file type */
  Elf64_Half	e_machine;		/* Architecture */
  Elf64_Word	e_version;		/* Object file version */
  Elf64_Addr	e_entry;		/* Entry point virtual address */
  Elf64_Off	e_phoff;		/* Program header table file offset */
  Elf64_Off	e_shoff;		/* Section header table file offset */
  Elf64_Word	e_flags;		/* Processor-specific flags */
  Elf64_Half	e_ehsize;		/* ELF header size in bytes */
  Elf64_Half	e_phentsize;		/* Program header table entry size */
  Elf64_Half	e_phnum;		/* Program header table entry count */
  Elf64_Half	e_shentsize;		/* Section header table entry size */
  Elf64_Half	e_shnum;		/* Section header table entry count */
  Elf64_Half	e_shstrndx;		/* Section header string table index */
} Elf64_Ehdr;

typedef struct
{
  Elf64_Word	p_type;			/* Segment type */
  Elf64_Word	p_flags;		/* Segment flags */
  Elf64_Off	p_offset;		/* Segment file offset */
  Elf64_Addr	p_vaddr;		/* Segment virtual address */
  Elf64_Addr	p_paddr;		/* Segment physical address */
  Elf64_Xword	p_filesz;		/* Segment size in file */
  Elf64_Xword	p_memsz;		/* Segment size in memory */
  Elf64_Xword	p_align;		/* Segment alignment */
} Elf64_Phdr;


#define OF_DT_BEGIN_NODE 0x1
#define OF_DT_END_NODE 0x2
#define OF_DT_PROP 0x3
#define OF_DT_END 0x9

struct devtree_header {
  uint32_t magic, totalsize, off_struct, off_strings, off_reserve;
  uint32_t version, last_comp, boot_cpuid, size_strings, pad;
};

static unsigned long compute_memsize(void)
{
  long r;
  unsigned long v1, v2, memsize;
  unsigned long puid, lpar_id;

  r = lv1_get_logical_pu_id(&puid);
  r = lv1_get_logical_partition_id(&lpar_id);
  r = lv1_get_repository_node_value(lpar_id,
				    0x0000000062690000,	/* "bi" */
				    0x7075000000000000,	/* "pu" */
				    puid,		/* #<puid> */
				    0x726d5f73697a6500,	/* "rm_size" */
				    &v1, &v2);
  if(r)
    memsize = 128*1024*1024;
  else
    memsize = v1;

  r = lv1_get_repository_node_value(lpar_id,
				    0x0000000062690000,	/* "bi" */
				    0x72676e746f74616c,	/* "rgntotal" */
				    0, 0,
				    &v1, &v2);
  if(!r && (v1-memsize)>2*(16*1024*1024))
    memsize += (v1-memsize-16*1024*1024)&~(16*1024*1024-1);

  return memsize;
}

static void create_devtree(void *addr, const char *cmdline)
{
  extern struct devtree_header dt_header;
  struct devtree_header *hdr = addr;
  uint8_t *dest, *src = (uint8_t *)&dt_header;
  uint32_t n, *ss;
  const char *name, *nodename="";
  *hdr = dt_header;
  dest = (uint8_t *)(hdr+1);

  hdr->off_reserve = dest-(uint8_t *)hdr;
  n = dt_header.off_struct - dt_header.off_reserve;
  memcpy(dest, src+dt_header.off_reserve, n);
  dest += n;

  hdr->off_struct = dest-(uint8_t *)hdr;
  ss = (uint32_t *)(src+dt_header.off_struct);
  while(*ss != OF_DT_END)
    switch(*ss++) {
    case OF_DT_BEGIN_NODE:
      *(uint32_t *)dest = OF_DT_BEGIN_NODE;
      dest += sizeof(uint32_t);
      nodename = (char *)ss;
      n = (strlen(nodename)+4)&~3;
      memcpy(dest, ss, n);
      dest += n;
      ss = (uint32_t *)(n+(char *)ss);
      break;
    case OF_DT_END_NODE:
      *(uint32_t *)dest = OF_DT_END_NODE;
      dest += sizeof(uint32_t);
      break;
    case OF_DT_PROP:
      n = *ss++;
      name = src+dt_header.off_strings+*ss;
      DBG("Processing property %s...", name);
      if(!strcmp(name, "reg") && !strcmp(nodename, "memory@0")) {
	unsigned long memsize = compute_memsize();
	((uint32_t *)dest)[0] = OF_DT_PROP;
	((uint32_t *)dest)[1] = 4*sizeof(uint32_t);
	((uint32_t *)dest)[2] = *ss;
	((uint32_t *)dest)[3] = 0;
	((uint32_t *)dest)[4] = 0;
	((uint32_t *)dest)[5] = (uint32_t)(memsize>>32);
	((uint32_t *)dest)[6] = (uint32_t)memsize;
	dest += sizeof(uint32_t)*7;
	DBG("Set to %lx\n", memsize);
      } else if(!strcmp(name, "linux,kernel-end")) {
	((uint32_t *)dest)[0] = OF_DT_PROP;
	((uint32_t *)dest)[1] = 2*sizeof(uint32_t);
	((uint32_t *)dest)[2] = *ss;
	((uint32_t *)dest)[3] = 0;
	((uint32_t *)dest)[4] = (uint32_t)kernel_memsize;
	dest += sizeof(uint32_t)*5;
	DBG("Set to %x\n", (uint32_t)kernel_memsize);
      } else if(!strcmp(name, "linux,initrd-end")) {
	if(initrd) {
	  ((uint32_t *)dest)[0] = OF_DT_PROP;
	  ((uint32_t *)dest)[1] = 2*sizeof(uint32_t);
	  ((uint32_t *)dest)[2] = *ss;
	  ((uint32_t *)dest)[3] = 0;
	  ((uint32_t *)dest)[4] = (uint32_t)(unsigned long)(initrd+initrd_size);
	  dest += sizeof(uint32_t)*5;
	  DBG("Set to %x\n", (uint32_t)(unsigned long)(initrd+initrd_size));
	} else {
	  DBG("Removed\n");
	}
      } else if(!strcmp(name, "linux,initrd-start")) {
	if(initrd) {
	  ((uint32_t *)dest)[0] = OF_DT_PROP;
	  ((uint32_t *)dest)[1] = 2*sizeof(uint32_t);
	  ((uint32_t *)dest)[2] = *ss;
	  ((uint32_t *)dest)[3] = 0;
	  ((uint32_t *)dest)[4] = (uint32_t)(unsigned long)initrd;
	  dest += sizeof(uint32_t)*5;
	  DBG("Set to %x\n", (uint32_t)(unsigned long)initrd);
	} else {
	  DBG("Removed\n");
	}
      } else if(!strcmp(name, "bootargs")) {
	((uint32_t *)dest)[0] = OF_DT_PROP;
	((uint32_t *)dest)[1] = strlen(cmdline)+1;
	((uint32_t *)dest)[2] = *ss;
	dest += sizeof(uint32_t)*3;
	memcpy(dest, cmdline, strlen(cmdline)+1);
	dest += strlen(cmdline)+1;
	while(3&(unsigned long)dest)
	  *dest++ = 0;
	DBG("Set to \"%s\"\n", cmdline);
      } else {
	((uint32_t *)dest)[0] = OF_DT_PROP;
	((uint32_t *)dest)[1] = n;
	dest += sizeof(uint32_t)*2;
	memcpy(dest, ss, (7+n)&~3);
	dest += (7+n)&~3;
	DBG("Copied\n");
      } 
      ss = (uint32_t *)(((7+n)&~3)+(char *)ss);	
    }
  *(uint32_t *)dest = OF_DT_END;
  dest += sizeof(uint32_t);

  hdr->off_strings = dest-(uint8_t *)hdr;
  memcpy(dest, src+dt_header.off_strings, dt_header.size_strings);
  dest += dt_header.size_strings;

  hdr->totalsize = dest-(uint8_t *)hdr;
}

static int load_initrd(struct file *file)
{
  initrd_size = (file->size+0xfff)&~0xfff;
  do
    initrd = memalign(0x1000, initrd_size);
  while(initrd && ((unsigned long)initrd) < kernel_memsize);
  if(!initrd) {
    fb_printf("Not enough memory to load initrd\n");
    return -1;
  }
  if(ext2_read(file, initrd, file->size) != file->size) {
    fb_printf("Failed to read initrd data\n");
    return -1;
  }
  if(file->size < initrd_size)
    memset(initrd+file->size, 0, initrd_size-file->size);

  return 0;
}

static int load_bld(struct file *file)
{
  char *cbuf = malloc(file->size);
  if(!cbuf) {
    fb_printf("Not enough memory to load compressed file\n");
    return -1;
  }

  /* These two bytes were already read by autodetection */
  cbuf[0] = 0x1f;
  cbuf[1] = 0x8b;

  if(ext2_read(file, cbuf+2, file->size-2) != file->size-2) {
    fb_printf("Failed to read compressed data\n");
    return -1;
  }
  unsigned int ucl = (cbuf[file->size-1]<<24)|(cbuf[file->size-2]<<16)|
    (cbuf[file->size-3]<<8)|cbuf[file->size-4];
  DBG("Compressed size: %ld, uncompressed size: %d\n", file->size, ucl);
  kernel = malloc(ucl);
  if(!kernel) {
    fb_printf("Not enough memory for uncompressed file\n");
    return -1;
  }

  fb_printf("Uncompressing image...\n");

  z_stream stream;
  int err;

  stream.next_in = (Bytef*)cbuf;
  stream.avail_in = (uInt)file->size;
  stream.next_out = kernel;
  stream.avail_out = (uInt)ucl;
  stream.zalloc = (alloc_func)0;
  stream.zfree = (free_func)0;
  err = inflateInit2(&stream, MAX_WBITS+32);
  if (err == Z_OK) {
    
    err = inflate(&stream, Z_FINISH);
    if (err == Z_STREAM_END) {
      kernel_size = stream.total_out;
      err = inflateEnd(&stream);
    } else
      inflateEnd(&stream);
    
  }
  
  free(cbuf);
  if(err != Z_OK) {
    fb_printf("zlib error %d on decompression\n", err);
    return -1;
  }

  return 1;
}

static int load_elf(struct file *file)
{
  Elf64_Ehdr ehdr;
  Elf64_Phdr phdr, phtmp;
  int ph, phfound=0;

  /* These two bytes were already read by autodetection */
  ehdr.e_ident[0] = 0x7f;
  ehdr.e_ident[1] = 'E';

  if(ext2_read(file, ((char *)&ehdr)+2, sizeof(ehdr)-2) != sizeof(ehdr)-2) {
    fb_printf("Failed to read ELF header\n");
    return -1;
  }
  
  if(memcmp(&ehdr, "\177ELF\2\2\1", 6)) {
    fb_printf("Invalid ELF header\n");
    return -1;
  }

  if(ehdr.e_type != 2 || ehdr.e_machine != 21 || ehdr.e_phentsize < sizeof(Elf64_Phdr)) {
    fb_printf("Unsupported ELF\n");
    return -1;
  }

  for(ph = 0; ph < ehdr.e_phnum; ph++) {
    file->offset = ehdr.e_phoff + ph * ehdr.e_phentsize;
    if(ext2_read(file, (char *)&phtmp, sizeof(phtmp)) != sizeof(phtmp)) {
      fb_printf("Failed to read program header\n");
      return -1;
    }
    if(phtmp.p_type == 1) {
      phdr = phtmp;
      phfound++;
    }
  }

  if(phfound != 1) {
    fb_printf("Error: Expected 1 loadable segment, found %d\n", phfound);
    return -1;
  }

  if(phdr.p_memsz > 128*1024*1024 || phdr.p_align > 128*1024*1024)
    kernel = NULL;
  else do
    kernel = memalign(phdr.p_align, phdr.p_memsz);
  while(kernel && ((unsigned long)kernel) < phdr.p_memsz);
  if(!kernel) {
    fb_printf("Not enough memory to load image\n");
    return -1;
  }

  file->offset = phdr.p_offset;
  if(ext2_read(file, kernel, phdr.p_filesz) != phdr.p_filesz) {
    fb_printf("Failed to read image data\n");
    return -1;
  }
  if(phdr.p_memsz > phdr.p_filesz)
    memset(kernel+phdr.p_filesz, 0, phdr.p_memsz-phdr.p_filesz);
  kernel_size = phdr.p_filesz; /* BSS not included */
  kernel_memsize = (phdr.p_memsz+0xfff)&~0xfff;

  return 2;
}

static int load_file(struct file *file)
{
  char buf[2];
  if(ext2_read(file, buf, 2) != 2) {
    fb_printf("\nFailed to read header\n");
    return -1;
  }
  if(file->size >= 10 && buf[0] == 0x1f && buf[1] == 0x8b) {
    DBG("File is BLD\n");
    return load_bld(file);
  } else if(file->size >= 16 && buf[0] == 0x7f && buf[1] == 'E') {
    DBG("File is ELF\n");
    return load_elf(file);
  } else {
    fb_printf("\nUnrecognized file type\n");
    return -1;
  }
}

static int parse_partition_id(char *id)
{
  int part, l = strlen(id);
  char *end;
  if(l<4 || memcmp(id, "sda", 3) || (part = strtol(id+3, &end, 10))<0 ||
     end != id+l) {
    fb_printf("\nUnsupported partition designation \"%s\"\n", id);
    return -1;
  }
  part = hdd_get_dos_partition(part-1);
  if(part < 0)
    fb_printf("\nPartition \"%s\" not found.\n", id);
  return part;
}

static int locate_file(struct file *file, const char *path, const char **p_end)
{
  int i, l, rc, part;
  const char *end;
  char *p;
  for(end=path; *end && *end != ' ' && *end != '\t' && *end != '\r'; end++)
    ;
  l = end-path;
  while(*end == ' ' || *end == '\t' || *end == '\r')
    end++;
  *p_end = end;
  p = malloc(l+1);
  if(!p) {
    DBG("OUT OF MEMORY!\n");
    return -1;
  }
  memcpy(p, path, l);
  p[l] = 0;
  path = p;
  part = hdd_get_root_partition();
  for(i=0; i<l; i++)
    if(p[i] == ':') {
      p[i] = 0;
      path = p+i+1;
      part = parse_partition_id(p);
      if(part < 0) {
	free(p);
	return -1;
      }
      break;
    }
  rc = ext2_open(file, part, path);
  free(p);
  return rc;
}

int load_object(const char *cmdline)
{
  struct file file;
  const char *end;
  int rc;

  DBG("Loading image\n");
  if(locate_file(&file, cmdline, &end)<0) {
    fb_printf("\nUnable to open %.*s\n", end-cmdline, cmdline);
    return -1;
  }
  rc = load_file(&file);
  ext2_close(&file);
  if(rc<0)
    return rc;
  cmdline = end;
  if(rc == 2 && !memcmp(cmdline, "initrd=", 7)) {
    int rc2;
    DBG("Loading initrd\n");
    cmdline += 7;
    if(locate_file(&file, cmdline, &end)<0) {
      fb_printf("\nUnable to open initrd %.*s\n", end-cmdline, cmdline);
      return -1;
    }
    rc2 = load_initrd(&file);
    ext2_close(&file);
    if(rc2<0)
      return rc2;
    cmdline = end;
  }
  if(rc == 2)
    create_devtree((void *)DEVTREE_ADDR, cmdline);
  return prepare_execute(rc, kernel, kernel_size, initrd, initrd_size,
			 (void*)DEVTREE_ADDR, (void*)SCRATCHPAD_ADDR);
}
